"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchAndConsumeMissedVaas = exports.nextVaaFetchingWorker = exports.consumeEventWithMissedVaaDetection = void 0;
const wh = require("@certusone/wormhole-sdk");
const logHelper_1 = require("../helpers/logHelper");
const utils_1 = require("../utils/utils");
const grpcWebNodeHttpTransport = require("@improbable-eng/grpc-web-node-http-transport");
const listenerHarness_1 = require("./listenerHarness");
const eventHarness_1 = require("./eventHarness");
const config_1 = require("../config");
const DEFAULT_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS = 5 * utils_1.minute;
const MIN_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS = 5 * utils_1.second;
let _logger;
const logger = () => {
    if (!_logger) {
        _logger = (0, logHelper_1.getScopedLogger)(["missedVaaFetching"]);
    }
    return _logger;
};
async function consumeEventWithMissedVaaDetection(vaa, plugin, storage, providers, extraData) {
    const parsedVaa = (0, utils_1.parseVaaWithBytes)(vaa);
    const chainId = (0, utils_1.assertChainId)(parsedVaa.emitterChain);
    const emitterAddress = (0, utils_1.wormholeBytesToHex)(parsedVaa.emitterAddress);
    const emitterRecord = await storage.getEmitterRecord(plugin.pluginName, chainId, emitterAddress);
    if (emitterRecord) {
        await fetchAndConsumeMissedVaas(plugin, storage, providers, chainId, emitterAddress, emitterRecord.lastSeenSequence, Number(parsedVaa.sequence));
    }
    await (0, eventHarness_1.consumeEventHarness)(vaa, plugin, storage, providers, extraData);
}
exports.consumeEventWithMissedVaaDetection = consumeEventWithMissedVaaDetection;
async function nextVaaFetchingWorker(plugins, storage, providers) {
    logger().debug(`Grouping emitter keys from plugins...`);
    const pluginsAndFilters = await getEmitterKeys(plugins);
    logger().debug(`Starting nextVaaFetchingWorker...`);
    const timeout = getTimeout();
    while (true) {
        logger().debug(`Pessimistically fetching next vaa for all emitters registered by plugins`);
        for (const key of pluginsAndFilters) {
            await tryFetchVaasForEmitter(key, storage, providers);
        }
        // wait 5 minutes between fetching next vaa
        logger().debug(`nextVaaFetchingWorker loop completed, sleeping ${timeout} ms...`);
        await (0, utils_1.sleep)(timeout);
    }
}
exports.nextVaaFetchingWorker = nextVaaFetchingWorker;
async function tryFetchVaasForEmitter({ plugin, chainId, emitterAddress, }, storage, providers) {
    const record = await storage.getEmitterRecord(plugin.pluginName, chainId, emitterAddress);
    if (record !== null) {
        let seq = record.lastSeenSequence + 1;
        for (let fetched = true; fetched; seq++) {
            fetched = await tryFetchAndConsumeVaa(plugin, storage, providers, chainId, emitterAddress, seq);
            if (fetched) {
                logger().info(`Found vaa past last known sequence number, now fetching ${seq + 1}`);
            }
        }
    }
}
async function tryFetchAndConsumeVaa(plugin, storage, providers, chainId, emitterAddress, sequence) {
    try {
        const resp = await wh.getSignedVAAWithRetry([(0, config_1.getCommonEnv)().wormholeRpc], chainId, emitterAddress, sequence.toString(), { transport: grpcWebNodeHttpTransport.NodeHttpTransport() });
        if (!resp?.vaaBytes) {
            logger().debug(`Attempted to fetch VAA but not found in wormhole rpc. ${chainId}:${emitterAddress}:${sequence.toString()}`);
            return false;
        }
        (0, eventHarness_1.consumeEventHarness)(resp.vaaBytes, plugin, storage, providers);
        return true;
    }
    catch (e) {
        logger().error("Attempted to fetch VAA but encountered error");
        logger().error(e);
        return false;
    }
}
async function fetchAndConsumeMissedVaas(plugin, storage, providers, chainId, emitterAddress, lastSeenSequence, latestSequence) {
    logger().info(`Fetching missed vaas for ${chainId}:${emitterAddress}, from ${lastSeenSequence} to ${latestSequence}`);
    for (let seq = lastSeenSequence + 1; seq < latestSequence; ++seq) {
        const fetched = await tryFetchAndConsumeVaa(plugin, storage, providers, chainId, emitterAddress, seq);
        if (!fetched) {
            logger().warn(`Failed to fetch missed vaa ${chainId}:${emitterAddress}:${seq.toString()}`);
        }
    }
}
exports.fetchAndConsumeMissedVaas = fetchAndConsumeMissedVaas;
async function getEmitterKeys(plugins) {
    return await Promise.all(plugins.map(async (plugin) => {
        const rawFilters = plugin.getFilters();
        return await Promise.all(rawFilters.map(async (x) => {
            return {
                ...(await (0, listenerHarness_1.transformEmitterFilter)(x)),
                plugin,
            };
        }));
    })).then(x => x.flatMap(y => y));
}
function getTimeout() {
    const listenerEnv = (0, config_1.getListenerEnv)();
    if (listenerEnv.nextVaaFetchingWorkerTimeoutSeconds) {
        const configValue = listenerEnv.nextVaaFetchingWorkerTimeoutSeconds * utils_1.second;
        if (configValue >= MIN_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS) {
            return configValue;
        }
        logger().warn(`nextVaaFetchingWorkerTimeoutSeconds must be at least ${Math.round(MIN_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS / utils_1.second)} seconds, found: ${listenerEnv.nextVaaFetchingWorkerTimeoutSeconds}`);
        return MIN_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS;
    }
    return DEFAULT_NEXT_VAA_FETCHING_WORKER_TIMEOUT_MS * utils_1.second;
}
//# sourceMappingURL=missedVaaFetching.js.map