"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSpyEventSource = void 0;
const wormhole_spydk_1 = require("@certusone/wormhole-spydk");
const LRUCache = require("lru-cache");
const utils_1 = require("../utils/utils");
const wormholeSdk = require("@certusone/wormhole-sdk");
const logHelper_1 = require("../helpers/logHelper");
const config_1 = require("../config");
const listenerHarness_1 = require("./listenerHarness");
const missedVaaFetching_1 = require("./missedVaaFetching");
// TODO: get from config or sdk etc.
const NUM_GUARDIANS = 19;
let _logger;
const logger = () => {
    if (!_logger) {
        _logger = (0, logHelper_1.getScopedLogger)(["spyEventSource"]);
    }
    return _logger;
};
async function createSpyEventSource(plugins, storage, providers) {
    const commonEnv = (0, config_1.getCommonEnv)();
    const listenerEnv = (0, config_1.getListenerEnv)();
    const spyClient = (0, wormhole_spydk_1.createSpyRPCServiceClient)(listenerEnv.spyServiceHost || "");
    plugins.forEach(plugin => {
        if (plugin.shouldSpy) {
            logger().info(`Initializing spy listener for plugin ${plugin.pluginName}...`);
            runPluginSpyListener(plugin, storage, spyClient, providers, commonEnv.numGuardians || NUM_GUARDIANS);
        }
    });
}
exports.createSpyEventSource = createSpyEventSource;
//used for both rest & spy relayer for now
async function runPluginSpyListener(plugin, storage, client, providers, numGuardians) {
    const vaaHashCache = new LRUCache({
        max: 10000,
    });
    while (true) {
        let stream;
        try {
            const rawFilters = plugin.getFilters();
            const filters = await Promise.all(rawFilters.map(async (x) => {
                return {
                    emitterFilter: await (0, listenerHarness_1.transformEmitterFilter)(x),
                };
            }));
            logger().info(`${plugin.pluginName} subscribing to spy with raw filters: ${JSON.stringify(rawFilters)}`);
            logger().debug(`${plugin.pluginName} using transformed filters: ${JSON.stringify(filters)}`);
            stream = await (0, wormhole_spydk_1.subscribeSignedVAA)(client, {
                filters,
            });
            stream.on("data", (vaa) => {
                const parsed = wormholeSdk.parseVaa(vaa.vaaBytes);
                const hash = parsed.hash.toString("base64");
                logger().debug(hash);
                logger().debug(parsed.emitterChain);
                if (parsed.guardianSignatures.length < Math.ceil((numGuardians * 2) / 3)) {
                    logger().debug(`Encountered VAA without enough signatures: ${parsed.guardianSignatures.length}, need ${Math.ceil((numGuardians * 2) / 3)}, ${hash}`);
                    return;
                }
                if (vaaHashCache.get(hash)) {
                    logger().debug(`Duplicate founds for hash ${hash}`);
                    return;
                }
                vaaHashCache.set(hash, true);
                (0, missedVaaFetching_1.consumeEventWithMissedVaaDetection)(vaa.vaaBytes, plugin, storage, providers);
            });
            let connected = true;
            stream.on("error", (err) => {
                logger().error("spy service returned an error: %o", err);
                connected = false;
            });
            stream.on("close", () => {
                logger().error("spy service closed the connection!");
                connected = false;
            });
            logger().info("connected to spy service, listening for transfer signed VAAs");
            while (connected) {
                await (0, utils_1.sleep)(1000);
            }
        }
        catch (e) {
            logger().error("spy service threw an exception: %o", e);
        }
        stream.destroy();
        await (0, utils_1.sleep)(5 * 1000);
        logger().info("attempting to reconnect to the spy service");
    }
}
//# sourceMappingURL=spyEventSource.js.map